import pytest
from sqlalchemy import select, text

from flask_audit_logger.expressions import jsonb_change_key_name
from tests.defaults.flask_app import db


@pytest.mark.usefixtures("test_client")
class TestJSONBChangeKeyName:
    @pytest.mark.parametrize(
        ("data", "old_key", "new_key", "expected"),
        (
            ('{"key1": 4, "key2": 3}', "key1", "key3", {"key3": 4, "key2": 3}),
            ('{"key1": 4}', "key2", "key3", {"key1": 4}),
            ('{"key1": 4, "key2": 3}', "key1", "key2", {"key2": 3}),
        ),
    )
    def test_raw_sql(self, data, old_key, new_key, expected):
        result = db.session.scalar(
            text(
                """SELECT jsonb_change_key_name(
                    '{data}'::jsonb,
                    '{old_key}',
                    '{new_key}'
                )""".format(
                    data=data, old_key=old_key, new_key=new_key
                )
            )
        )
        assert result == expected

    @pytest.mark.parametrize(
        ("data", "old_key", "new_key", "expected"),
        (
            ({"key1": 4, "key2": 3}, "key1", "key3", {"key3": 4, "key2": 3}),
            ({"key1": 4}, "key2", "key3", {"key1": 4}),
            ({"key1": 4, "key2": 3}, "key1", "key2", {"key2": 3}),
        ),
    )
    def test_sqlalchemy_function_expr(self, data, old_key, new_key, expected):
        result = db.session.scalar(select(jsonb_change_key_name(data, old_key, new_key)))
        assert result == expected



import os
import shutil
from contextlib import contextmanager, redirect_stdout
from io import StringIO
from pathlib import Path
from typing import Any, Callable, NoReturn

from alembic import command as alem_command
from alembic.config import Config
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import Engine, text

REPO_ROOT = Path(os.path.abspath(os.path.dirname(__file__))).parent.resolve()

ALEMBIC_COMMAND_MAP: dict[str, Callable[..., NoReturn]] = {
    "upgrade": alem_command.upgrade,
    "downgrade": alem_command.downgrade,
    "revision": alem_command.revision,
}


@contextmanager
def run_audit_logger_migrations(db: SQLAlchemy, alembic_config: Path):
    clear_alembic_migrations(db, alembic_config)
    run_alembic_command(
        engine=db.engine,
        command="revision",
        command_kwargs={"autogenerate": True, "rev_id": "1", "message": "create"},
        alembic_config=alembic_config,
    )

    run_alembic_command(
        engine=db.engine,
        command="upgrade",
        command_kwargs={"revision": "head"},
        alembic_config=alembic_config,
    )

    yield

    run_alembic_command(
        engine=db.engine,
        command="downgrade",
        command_kwargs={"revision": "base"},
        alembic_config=alembic_config,
    )
    clear_alembic_migrations(db, alembic_config)


def run_alembic_command(
    engine: Engine, command: str, command_kwargs: dict[str, Any], alembic_config: Path
) -> str:
    command_func = ALEMBIC_COMMAND_MAP[command]

    stdout = StringIO()

    alembic_cfg = Config(alembic_config / "alembic.ini")

    # Make double sure alembic references the test database
    alembic_cfg.set_main_option("sqlalchemy.url", engine.url.render_as_string(hide_password=False))
    alembic_cfg.set_main_option("script_location", str(alembic_config))

    with engine.begin() as connection:
        alembic_cfg.attributes["connection"] = connection
        with redirect_stdout(stdout):
            command_func(alembic_cfg, **command_kwargs)
    return stdout.getvalue()


def clear_alembic_migrations(db, alembic_config):
    with db.engine.begin() as connection:
        connection.execute(text("DROP SCHEMA public CASCADE; CREATE SCHEMA public;"))
        connection.execute(
            text("DROP SCHEMA IF EXISTS audit_logs CASCADE; CREATE SCHEMA audit_logs;")
        )

    versions_root = alembic_config / "versions"

    # Remove any migrations that were left behind
    versions_root.mkdir(exist_ok=True, parents=True)
    shutil.rmtree(versions_root)
    versions_root.mkdir(exist_ok=True, parents=True)



import pytest
from sqlalchemy import Text, cast, select

from flask_audit_logger.migrations import (
    add_column,
    alter_column,
    change_column_name,
    remove_column,
    rename_table,
)
from tests.defaults.flask_app import AuditLogActivity, db


@pytest.mark.usefixtures("test_client")
class TestRenameTable:
    def test_only_updates_given_table(self, article, user):
        rename_table(db.session, "user", "user2")
        activity = db.session.scalar(
            select(AuditLogActivity).where(AuditLogActivity.table_name == "article").limit(1)
        )
        assert activity

    def test_updates_table_name(self, user):
        rename_table(db.session, "user", "user2")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.table_name == "user2"


@pytest.mark.usefixtures("test_client")
class TestChangeColumnName:
    def test_only_updates_given_table(self, article, user):
        change_column_name(db.session, "user", "name", "some_name")
        activity = db.session.scalar(
            select(AuditLogActivity).where(AuditLogActivity.table_name == "article").limit(1)
        )
        assert "name" in activity.changed_data

    def test_updates_changed_data(self, user):
        change_column_name(db.session, "user", "name", "some_name")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.changed_data == {
            "id": user.id,
            "some_name": "Jan",
            "age": 15,
        }

    def test_updates_old_data(self, user):
        user.name = "Luke"
        db.session.commit()
        change_column_name(db.session, "user", "name", "some_name")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.old_data == {"id": user.id, "some_name": "Jan", "age": 15}


@pytest.mark.usefixtures("test_client")
class TestRemoveColumn:
    def test_only_updates_given_table(self, article, user):
        remove_column(db.session, "user", "name")
        activity = db.session.scalar(
            select(AuditLogActivity).where(AuditLogActivity.table_name == "article").limit(1)
        )
        assert "name" in activity.changed_data

    def test_updates_changed_data(self, user):
        remove_column(db.session, "user", "name")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.old_data == {}
        assert activity.changed_data == {"id": user.id, "age": 15}

    def test_updates_old_data(self, user):
        user.name = "Luke"
        db.session.commit()
        remove_column(db.session, "user", "name")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.old_data == {"id": user.id, "age": 15}


@pytest.mark.usefixtures("test_client")
class TestAddColumn:
    def test_only_updates_given_table(self, article, user):
        add_column(db.session, "user", "some_column")
        activity = db.session.scalar(
            select(AuditLogActivity).where(AuditLogActivity.table_name == "article").limit(1)
        )
        assert "some_column" not in activity.changed_data

    def test_updates_changed_data(self, user):
        add_column(db.session, "user", "some_column")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.old_data == {}
        assert activity.changed_data == {
            "id": user.id,
            "age": 15,
            "name": "Jan",
            "some_column": None,
        }

    def test_updates_old_data(self, user):
        user.name = "Luke"
        db.session.commit()
        add_column(db.session, "user", "some_column")
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.old_data == {
            "id": user.id,
            "age": 15,
            "name": "Jan",
            "some_column": None,
        }
        assert activity.changed_data == {"name": "Luke"}


@pytest.mark.usefixtures("test_client")
class TestAlterColumn:
    def test_only_updates_given_table(self, article, user):
        alter_column(db.session, "user", "id", lambda value, activity_table: cast(value, Text))
        activity = db.session.scalar(
            select(AuditLogActivity).where(AuditLogActivity.table_name == "article").limit(1)
        )
        assert isinstance(activity.changed_data["id"], int)

    def test_updates_changed_data(self, user):
        alter_column(db.session, "user", "id", lambda value, activity_table: cast(value, Text))
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.changed_data == {
            "id": str(user.id),
            "age": 15,
            "name": "Jan",
        }

    def test_updates_old_data(self, user):
        user.name = "Luke"
        db.session.commit()
        alter_column(db.session, "user", "id", lambda value, activity_table: cast(value, Text))
        activity = db.session.scalar(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at.desc()).limit(1)
        )
        assert activity.old_data == {"id": str(user.id), "age": 15, "name": "Jan"}



import re
from collections import defaultdict, namedtuple

from alembic.autogenerate import comparators, renderers
from alembic.operations import MigrateOperation, Operations
from sqlalchemy import text


def setup_functions_and_triggers(audit_logger):
    """The comparator at the end controls which functions and triggers
    need to be added/removed based on the current audit_logger.versioned_tables."""

    """ PG Functions """

    @Operations.register_operation("init_audit_logger_function")
    class InitAuditLoggerFunctionOp(MigrateOperation):
        def __init__(self, function_signature):
            self.function_signature = function_signature

        @classmethod
        def init_audit_logger_function(cls, operations, function_signature):
            op = InitAuditLoggerFunctionOp(function_signature)
            return operations.invoke(op)

        def reverse(self):
            return RemoveAuditLoggerFunctionOp(self.function_signature)

    @Operations.register_operation("remove_audit_logger_function")
    class RemoveAuditLoggerFunctionOp(MigrateOperation):
        def __init__(self, function_signature):
            self.function_signature = function_signature

        @classmethod
        def remove_audit_logger_function(cls, operations, function_signature):
            op = RemoveAuditLoggerFunctionOp(function_signature)
            return operations.invoke(op)

        def reverse(self):
            return InitAuditLoggerFunctionOp(self.function_signature)

    @Operations.implementation_for(InitAuditLoggerFunctionOp)
    def init_audit_logger_function(operations, operation):
        pg_func = audit_logger.functions_by_signature[operation.function_signature]
        operations.execute(pg_func.create_sql)

    @Operations.implementation_for(RemoveAuditLoggerFunctionOp)
    def remove_audit_logger_function(operations, operation):
        pg_func = audit_logger.functions_by_signature[operation.function_signature]
        operations.execute(pg_func.drop_sql)

    @renderers.dispatch_for(InitAuditLoggerFunctionOp)
    def render_init_audit_logger_function(autogen_context, op):
        return "op.init_audit_logger_function(%r)" % (op.function_signature,)

    @renderers.dispatch_for(RemoveAuditLoggerFunctionOp)
    def render_remove_audit_logger_function(autogen_context, op):
        return "op.remove_audit_logger_function(%r)" % (op.function_signature,)

    """ Table Triggers """

    @Operations.register_operation("init_audit_logger_triggers")
    class InitAuditLoggerTriggers(MigrateOperation):
        def __init__(
            self,
            table_name,
            excluded_columns=None,
            original_excluded_columns=None,
        ):
            self.table_name = table_name
            self.excluded_columns = excluded_columns or []
            self.original_excluded_columns = original_excluded_columns or []

        @classmethod
        def init_audit_logger_triggers(cls, operations, table_name, excluded_columns=None):
            op = InitAuditLoggerTriggers(table_name, excluded_columns=excluded_columns)
            return operations.invoke(op)

        def reverse(self):
            return RemoveAuditLoggerTriggers(
                self.table_name, excluded_columns=self.original_excluded_columns
            )

    @Operations.register_operation("remove_audit_logger_triggers")
    class RemoveAuditLoggerTriggers(MigrateOperation):
        def __init__(self, table_name, excluded_columns=None):
            self.table_name = table_name
            self.excluded_columns = excluded_columns or []

        @classmethod
        def remove_audit_logger_triggers(cls, operations, table_name, **kwargs):
            op = RemoveAuditLoggerTriggers(table_name, **kwargs)
            return operations.invoke(op)

        def reverse(self):
            return InitAuditLoggerTriggers(self.table_name, excluded_columns=self.excluded_columns)

    @Operations.implementation_for(InitAuditLoggerTriggers)
    def init_audit_logger_triggers(operations, operation):
        pg_triggers = audit_logger.pg_triggers_per_table.get(operation.table_name, [])
        for pg_trigger in pg_triggers:
            operations.execute(pg_trigger.create_sql)

    @Operations.implementation_for(RemoveAuditLoggerTriggers)
    def remove_audit_logger_triggers(operations, operation):
        pg_triggers = audit_logger.pg_triggers_per_table.get(operation.table_name, [])
        for pg_trigger in pg_triggers:
            operations.execute(pg_trigger.drop_sql)

    @renderers.dispatch_for(InitAuditLoggerTriggers)
    def render_init_audit_logger_triggers(autogen_context, op):
        if not op.excluded_columns:
            return "op.init_audit_logger_triggers(%r)" % (op.table_name,)
        return "op.init_audit_logger_triggers(%r, excluded_columns=%r)" % (
            op.table_name,
            op.excluded_columns,
        )

    @renderers.dispatch_for(RemoveAuditLoggerTriggers)
    def render_remove_audit_logger_triggers(autogen_context, op):
        return "op.remove_audit_logger_triggers(%r)" % (op.table_name)

    @comparators.dispatch_for("schema")
    def check_for_function_and_trigger_changes(autogen_context, upgrade_ops, schemas):
        check_functions(autogen_context, upgrade_ops, schemas)
        check_triggers(autogen_context, upgrade_ops, schemas)

    def check_functions(autogen_context, upgrade_ops, schemas):
        funcs_sql = text(
            f"""
            SELECT 
                format(
                    '%%s(%%s)',
                    replace(p.oid::regproc::TEXT, '{audit_logger.prefix}', ''),
                    pg_get_function_identity_arguments(p.oid)
                ) AS signature
            FROM pg_catalog.pg_proc p
            JOIN pg_catalog.pg_namespace n ON n.oid = p.pronamespace
            WHERE n.nspname = '{audit_logger.schema}'
            ORDER  BY 1;
        """
        )

        funcs_in_db = {row.signature for row in autogen_context.connection.execute(funcs_sql)}

        funcs_in_app = set(audit_logger.functions_by_signature.keys())

        should_add = funcs_in_app - funcs_in_db
        should_remove = funcs_in_db - funcs_in_app

        for func in audit_logger.pg_functions:
            sig = func.signature
            if sig in should_add:
                upgrade_ops.ops.append(InitAuditLoggerFunctionOp(sig))
            if sig in should_remove:
                upgrade_ops.ops.append(RemoveAuditLoggerFunctionOp(sig))

    Trigger = namedtuple("Trigger", ["table", "trigger_name", "definition"])

    def check_triggers(autogen_context, upgrade_ops, schemas):
        triggers_sql = text(
            """
            SELECT event_object_table, trigger_name, action_statement
            FROM information_schema.triggers
            WHERE trigger_name LIKE 'audit_trigger%%'
        """
        )

        triggers_per_table = defaultdict(list)
        for row in autogen_context.connection.execute(triggers_sql):
            trigger = Trigger(*row)
            triggers_per_table[trigger.table].append(trigger)

        for table in audit_logger.versioned_tables:
            trigger = next((tr for tr in triggers_per_table[table.name]), None)
            excluded_columns_in_app = table.info["versioned"].get("exclude", [])

            if not trigger:
                upgrade_ops.ops.append(
                    InitAuditLoggerTriggers(
                        table.name,
                        excluded_columns=excluded_columns_in_app,
                    )
                )
            else:
                excluded_columns_in_db = _get_existing_excluded_columns(trigger.definition)

                if set(excluded_columns_in_db) != set(excluded_columns_in_app):
                    upgrade_ops.ops.append(
                        InitAuditLoggerTriggers(
                            table.name,
                            excluded_columns=excluded_columns_in_app,
                            original_excluded_columns=excluded_columns_in_db,
                        )
                    )

        for table_name, triggers in triggers_per_table.items():
            if table_name not in {t.name for t in audit_logger.versioned_tables}:
                upgrade_ops.ops.append(
                    RemoveAuditLoggerTriggers(
                        table_name,
                        excluded_columns=_get_existing_excluded_columns(triggers[0].definition),
                    )
                )


def _get_existing_excluded_columns(trigger_definition):
    matched = re.search(r"create_activity\('{(.+)}'\)", trigger_definition)

    if matched:
        return matched.group(1).split(",")

    return []



from alembic.autogenerate import renderers, rewriter
from alembic.operations import MigrateOperation, Operations, ops

from flask_audit_logger.migrations import add_column, remove_column


def init_migration_ops(schema: str):
    @Operations.register_operation("add_column_to_activity")
    class AddColumnToActivityOp(MigrateOperation):
        """Initialize Activity Table Triggers"""

        def __init__(self, table_name, column_name, default_value=None):
            self.table_name = table_name
            self.column_name = column_name
            self.default_value = default_value

        @classmethod
        def add_column_to_activity(cls, operations, table_name, column_name, **kwargs):
            op = AddColumnToActivityOp(table_name, column_name, **kwargs)
            return operations.invoke(op)

        def reverse(self):
            # only needed to support autogenerate
            return RemoveColumnFromRemoveActivityOp(
                self.table_name,
                self.column_name,
                default_value=self.default_value,
            )

    @Operations.register_operation("remove_column_from_activity")
    class RemoveColumnFromRemoveActivityOp(MigrateOperation):
        """Drop Activity Table Triggers"""

        def __init__(self, table_name, column_name, default_value=None):
            self.table_name = table_name
            self.column_name = column_name
            self.default_value = default_value

        @classmethod
        def remove_column_from_activity(cls, operations, table_name, column_name, **kwargs):
            op = RemoveColumnFromRemoveActivityOp(table_name, column_name, **kwargs)
            return operations.invoke(op)

        def reverse(self):
            # only needed to support autogenerate
            return AddColumnToActivityOp(
                self.table_name,
                self.column_name,
                default_value=self.default_value,
            )

    @Operations.implementation_for(AddColumnToActivityOp)
    def add_column_to_activity(operations, operation):
        add_column(
            operations,
            operation.table_name,
            operation.column_name,
            default_value=operation.default_value,
            schema=schema,
        )

    @Operations.implementation_for(RemoveColumnFromRemoveActivityOp)
    def remove_column_from_activity(operations, operation):
        conn = operations.connection
        remove_column(conn, operation.table_name, operation.column_name, schema)

    @renderers.dispatch_for(AddColumnToActivityOp)
    def render_add_column_to_activity(autogen_context, op):
        return "op.add_column_to_activity(%r, %r, default_value=%r)" % (
            op.table_name,
            op.column_name,
            op.default_value,
        )

    @renderers.dispatch_for(RemoveColumnFromRemoveActivityOp)
    def render_remove_column_from_activity(autogen_context, op):
        return "op.remove_column_from_activity(%r, %r)" % (
            op.table_name,
            op.column_name,
        )

    writer = rewriter.Rewriter()

    @writer.rewrites(ops.AddColumnOp)
    def add_column_rewrite(context, revision, op):
        table_info = op.column.table.info or {}
        if "versioned" in table_info and op.column.name not in table_info["versioned"].get(
            "exclude", []
        ):
            return [
                op,
                AddColumnToActivityOp(op.table_name, op.column.name),
            ]
        else:
            return op

    @writer.rewrites(ops.DropColumnOp)
    def drop_column_rewrite(context, revision, op):
        column = op.to_column()
        table_info = column.table.info or {}
        if "versioned" in table_info and column.name not in table_info["versioned"].get(
            "exclude", []
        ):
            return [
                op,
                RemoveColumnFromRemoveActivityOp(op.table_name, column.name),
            ]
        else:
            return op

    return writer



from .migration_ops import init_migration_ops
from .setup_functions_and_triggers import setup_functions_and_triggers
from .setup_schema import setup_schema

__all__ = [
    "setup_functions_and_triggers",
    "setup_schema",
    "init_migration_ops",
]



import os
import string
from contextlib import contextmanager
from dataclasses import dataclass
from functools import cached_property

from flask import request
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import (
    DDL,
    BigInteger,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    Table,
    Text,
    event,
    func,
    inspect,
    literal_column,
    text,
)
from sqlalchemy.dialects.postgresql import INET, JSONB, ExcludeConstraint, insert
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import ColumnProperty, relationship
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.elements import TextClause

from flask_audit_logger import alembic

HERE = os.path.dirname(os.path.abspath(__file__))


class ImproperlyConfigured(Exception):
    pass


@dataclass
class PGExtension:
    schema: str
    signature: str

    @property
    def create_sql(self):
        return text(f"CREATE EXTENSION IF NOT EXISTS {self.signature} WITH SCHEMA {self.schema}")

    @property
    def drop_sql(self):
        return text(f"DROP EXTENSION IF EXISTS {self.signature}")


@dataclass
class PGFunction:
    schema: str
    signature: str
    create_sql: TextClause

    @property
    def drop_sql(self):
        return text(f'DROP FUNCTION IF EXISTS "{self.schema}"."{self.signature}" CASCADE')


@dataclass
class PGTrigger:
    schema: str
    signature: str
    table_name: str
    create_sql: TextClause

    @property
    def drop_sql(self):
        return text(f'DROP TRIGGER IF EXISTS "{self.signature}" ON "{self.table_name}"')


class AuditLogger(object):
    _actor_cls = None
    writer = None

    def __init__(
        self,
        db,
        get_actor_id=None,
        get_client_addr=None,
        actor_cls=None,
        schema=None,
    ):
        self._actor_cls = actor_cls or "User"
        self.get_actor_id = get_actor_id or _default_actor_id
        self.get_client_addr = get_client_addr or _default_client_addr
        self.schema = schema or "public"
        self.audit_logger_disabled = False
        self.db = db
        self.transaction_cls = _transaction_model_factory(db.Model, schema, self.actor_cls)
        self.activity_cls = _activity_model_factory(db.Model, schema, self.transaction_cls)
        self.versioned_tables = _detect_versioned_tables(db)
        self.attach_listeners()
        self.initialize_alembic_operations()

    def attach_listeners(self):
        """Listeners save transaction records with actor_ids when versioned tables are affected.
        Flush events occur when a mapped object is created or modified. ORM Execute events occur
        when an insert()/update()/delete() is passed to session.execute()."""
        event.listen(Session, "before_flush", self.receive_before_flush)
        event.listen(Session, "do_orm_execute", self.receive_do_orm_execute)

    def initialize_alembic_operations(self):
        alembic.setup_schema(self)
        alembic.setup_functions_and_triggers(self)
        self.writer = alembic.init_migration_ops(self.schema)

    def process_revision_directives(self, context, revision, directives):
        if self.writer:
            self.writer.process_revision_directives(context, revision, directives)

    @property
    def prefix(self):
        return f"{self.schema}." if self.schema != "public" else ""

    @cached_property
    def pg_entities(self):
        return [
            self.pg_btree_gist_extension,
            *self.pg_functions,
            *self.pg_triggers,
        ]

    @cached_property
    def pg_functions(self):
        return [
            self.pg_get_setting,
            self.pg_jsonb_subtract,
            self.pg_jsonb_change_key_name,
            self.pg_create_activity,
        ]

    @property
    def pg_triggers(self):
        return [t for triggers in self.pg_triggers_per_table.values() for t in triggers]

    @cached_property
    def pg_triggers_per_table(self):
        triggers_per_table = {}
        for table in self.versioned_tables:
            target_schema = table.schema or "public"
            versioned = table.info.get("versioned", {})
            excluded_columns = ""
            if "exclude" in versioned:
                joined_excludes = ",".join(versioned["exclude"])
                excluded_columns = "'{" + joined_excludes + "}'"

            triggers_per_table[table.name] = [
                PGTrigger(
                    schema=target_schema,
                    table_name=table.name,
                    signature="audit_trigger_insert",
                    create_sql=self.render_sql_template(
                        "audit_trigger_insert.sql",
                        table_name=table.name,
                        excluded_columns=excluded_columns,
                    ),
                ),
                PGTrigger(
                    schema=target_schema,
                    table_name=table.name,
                    signature="audit_trigger_update",
                    create_sql=self.render_sql_template(
                        "audit_trigger_update.sql",
                        table_name=table.name,
                        excluded_columns=excluded_columns,
                    ),
                ),
                PGTrigger(
                    schema=target_schema,
                    table_name=table.name,
                    signature="audit_trigger_delete",
                    create_sql=self.render_sql_template(
                        "audit_trigger_delete.sql",
                        table_name=table.name,
                        excluded_columns=excluded_columns,
                    ),
                ),
            ]

        return triggers_per_table

    @property
    def functions_by_signature(self) -> dict[str, PGFunction]:
        return {pg_func.signature: pg_func for pg_func in self.pg_functions}

    @cached_property
    def pg_btree_gist_extension(self) -> PGExtension:
        return PGExtension(schema="public", signature="btree_gist")

    @property
    def pg_get_setting(self) -> PGFunction:
        return PGFunction(
            schema=self.schema,
            signature="get_setting(setting text, fallback text)",
            create_sql=self.render_sql_template("get_setting.sql"),
        )

    @property
    def pg_jsonb_subtract(self) -> PGFunction:
        return PGFunction(
            schema=self.schema,
            signature="jsonb_subtract(arg1 jsonb, arg2 jsonb)",
            create_sql=self.render_sql_template("jsonb_subtract.sql"),
        )

    @property
    def pg_jsonb_change_key_name(self) -> PGFunction:
        return PGFunction(
            schema=self.schema,
            signature="jsonb_change_key_name(data jsonb, old_key text, new_key text)",
            create_sql=self.render_sql_template("jsonb_change_key_name.sql"),
        )

    @property
    def pg_create_activity(self) -> PGFunction:
        return PGFunction(
            schema=self.schema,
            signature="create_activity()",
            create_sql=self.render_sql_template("create_activity.sql"),
        )

    @contextmanager
    def disable(self, session):
        session.execute(text("SET LOCAL flask_audit_logger.enable_versioning = 'false'"))
        self.audit_logger_disabled = True
        try:
            yield
        finally:
            self.audit_logger_disabled = False
            session.execute(text("SET LOCAL flask_audit_logger.enable_versioning = 'true'"))

    def render_sql_template(
        self, tmpl_name: str, as_text: bool = True, **kwargs
    ) -> TextClause | DDL:
        file_contents = _read_file(f"templates/{tmpl_name}").replace("$$", "$$$$")
        tmpl = string.Template(file_contents)
        context = dict(schema=self.schema)

        context["schema_prefix"] = "{}.".format(self.schema)
        context["revoke_cmd"] = ("REVOKE ALL ON {schema_prefix}activity FROM public;").format(
            **context
        )

        sql = tmpl.substitute(**context, **kwargs)

        if not as_text:
            return DDL(sql)

        return text(sql)

    def receive_do_orm_execute(self, orm_execute_state):
        is_write = (
            orm_execute_state.is_insert
            or orm_execute_state.is_update
            or orm_execute_state.is_delete
        )
        affects_versioned_table = any(
            m.local_table in self.versioned_tables for m in orm_execute_state.all_mappers
        )
        if is_write and affects_versioned_table:
            self.save_transaction(orm_execute_state.session)

    def receive_before_flush(self, session, flush_context, instances):
        if _is_session_modified(session, self.versioned_tables):
            self.save_transaction(session)

    def save_transaction(self, session):
        if self.audit_logger_disabled:
            return

        values = {
            "native_transaction_id": func.txid_current(),
            "issued_at": text("now() AT TIME ZONE 'UTC'"),
            "client_addr": self.get_client_addr(),
            "actor_id": self.get_actor_id(),
        }

        stmt = (
            insert(self.transaction_cls)
            .values(**values)
            .on_conflict_do_nothing(constraint="transaction_unique_native_tx_id")
        )
        session.execute(stmt)

    @property
    def actor_cls(self):
        if isinstance(self._actor_cls, str):
            if not self.db.Model:
                raise ImproperlyConfigured("No SQLAlchemy db object")
            registry = self.db.Model.registry._class_registry
            try:
                return registry[self._actor_cls]
            except KeyError:
                raise ImproperlyConfigured(
                    f"""Could not build relationship between AuditLogActivity
                    and {self._actor_cls}. {self._actor_cls} was not found in
                    declarative class registry. Either configure AuditLogger to
                    use different actor class or disable this relationship by
                    setting it to None."""
                )
        return self._actor_cls


def _transaction_model_factory(base, schema, actor_cls):
    if actor_cls:
        actor_pk = inspect(actor_cls).primary_key[0]
        actor_fk = ForeignKey(f"{actor_cls.__table__.name}.{actor_pk.name}")

    class AuditLogTransaction(base):
        __tablename__ = "transaction"

        id = Column(BigInteger, primary_key=True)
        native_transaction_id = Column(BigInteger)
        issued_at = Column(DateTime)
        client_addr = Column(INET)
        if actor_cls:
            actor_id = Column(actor_pk.type, actor_fk)
            actor = relationship(actor_cls)
        else:
            actor_id = Column(Text)

        __table_args__ = (
            ExcludeConstraint(
                (literal_column("native_transaction_id"), "="),
                (
                    literal_column("tsrange(issued_at - INTERVAL '1 HOUR', issued_at)"),
                    "&&",
                ),
                name="transaction_unique_native_tx_id",
            ),
            {"schema": schema},
        )

        def __repr__(self):
            return "<{cls} id={id!r} issued_at={issued_at!r}>".format(
                cls=self.__class__.__name__, id=self.id, issued_at=self.issued_at
            )

    return AuditLogTransaction


def _activity_model_factory(base, schema_name, transaction_cls):
    class AuditLogActivity(base):
        __tablename__ = "activity"
        __table_args__ = {"schema": schema_name}

        id = Column(BigInteger, primary_key=True)
        schema = Column(Text)
        table_name = Column(Text)
        relid = Column(Integer)
        issued_at = Column(DateTime)
        native_transaction_id = Column(BigInteger, index=True)
        verb = Column(Text)
        old_data = Column(JSONB, default={}, server_default="{}")
        changed_data = Column(JSONB, default={}, server_default="{}")
        transaction_id = Column(BigInteger, ForeignKey(transaction_cls.id))

        transaction = relationship(transaction_cls, backref="activities")

        @hybrid_property
        def data(self):
            data = self.old_data.copy() if self.old_data else {}
            if self.changed_data:
                data.update(self.changed_data)
            return data

        @data.expression
        def data(cls):
            return cls.old_data + cls.changed_data

        def __repr__(self):
            return ("<{cls} table_name={table_name!r} " "id={id!r}>").format(
                cls=self.__class__.__name__, table_name=self.table_name, id=self.id
            )

    return AuditLogActivity


def _read_file(file):
    with open(os.path.join(HERE, file)) as f:
        s = f.read()
    return s


def _default_actor_id():
    try:
        from flask_login import current_user
    except ImportError:
        return None

    try:
        return current_user.id
    except AttributeError:
        return None


def _default_client_addr():
    # Return None if we are outside of request context.
    return (request and request.remote_addr) or None


def _detect_versioned_tables(db: SQLAlchemy) -> set[Table]:
    versioned_tables = set()

    for table in db.metadata.tables.values():
        if table.info.get("versioned") is not None:
            versioned_tables.add(table)

    return versioned_tables


def _is_session_modified(session: Session, versioned_tables: set[Table]) -> bool:
    return any(
        _is_entity_modified(entity) or entity in session.deleted
        for entity in session
        if entity.__table__ in versioned_tables
    )


def _is_entity_modified(entity) -> bool:
    versioned = entity.__table__.info.get("versioned")
    excluded_cols = set(versioned.get("exclude", []))
    modified_cols = {column.name for column in _modified_columns(entity)}

    return bool(modified_cols - excluded_cols)


def _modified_columns(obj):
    columns = set()
    mapper = inspect(obj.__class__)
    for key, attr in inspect(obj).attrs.items():
        if key in mapper.synonyms.keys():
            continue
        prop = getattr(obj.__class__, key).property
        if attr.history.has_changes():
            columns |= set(
                prop.columns
                if isinstance(prop, ColumnProperty)
                else [local for local, remote in prop.local_remote_pairs]
            )

    return columns



import sqlalchemy as sa
from sqlalchemy import update
from sqlalchemy.dialects.postgresql import JSONB

from .expressions import jsonb_change_key_name


def get_activity_table(schema=None):
    return sa.Table(
        "activity",
        sa.MetaData(),
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("table_name", sa.String),
        sa.Column("verb", sa.String),
        sa.Column("old_data", JSONB),
        sa.Column("changed_data", JSONB),
        schema=schema,
    )


def alter_column(conn, table, column_name, func, schema=None):
    """
    Run given callable against given table and given column in activity table
    jsonb data columns. This function is useful when you want to reflect type
    changes in your schema to activity table.

    In the following example we change the data type of User's age column from
    string to integer.


    ::

        from alembic import op
        from flask_audit_logger import alter_column


        def upgrade():
            op.alter_column(
                'user',
                'age',
                type_=sa.Integer
            )

            alter_column(
                op,
                'user',
                'age',
                lambda value, activity_table: sa.cast(value, sa.Integer)
            )


    :param conn:
        An object that is able to execute SQL (either SQLAlchemy Connection,
        Engine or Alembic Operations object)
    :param table:
        The table to run the column name changes against
    :param column_name:
        Name of the column to run callable against
    :param func:
        A callable to run against specific column in activity table jsonb data
        columns. The callable should take two parameters the jsonb value
        corresponding to given column_name and activity table object.
    :param schema:
        Optional name of schema to use.
    """
    activity_table = get_activity_table(schema=schema)
    query = (
        update(activity_table)
        .values(
            old_data=(
                activity_table.c.old_data
                + sa.cast(
                    sa.func.json_build_object(
                        column_name,
                        func(activity_table.c.old_data[column_name], activity_table),
                    ),
                    JSONB,
                )
            ),
            changed_data=(
                activity_table.c.changed_data
                + sa.cast(
                    sa.func.json_build_object(
                        column_name,
                        func(activity_table.c.changed_data[column_name], activity_table),
                    ),
                    JSONB,
                )
            ),
        )
        .where(activity_table.c.table_name == table)
    )
    return conn.execute(query)


def change_column_name(conn, table, old_column_name, new_column_name, schema=None):
    """
    Changes given `activity` jsonb data column key. This function is useful
    when you want to reflect column name changes to activity table.

    ::

        from alembic import op
        from flask_audit_logger import change_column_name


        def upgrade():
            op.alter_column(
                'my_table',
                'my_column',
                new_column_name='some_column'
            )

            change_column_name(op, 'my_table', 'my_column', 'some_column')


    :param conn:
        An object that is able to execute SQL (either SQLAlchemy Connection,
        Engine or Alembic Operations object)
    :param table:
        The table to run the column name changes against
    :param old_column_name:
        Name of the column to change
    :param new_column_name:
        New colum name
    :param schema:
        Optional name of schema to use.
    """
    activity_table = get_activity_table(schema=schema)
    query = (
        update(activity_table)
        .values(
            old_data=jsonb_change_key_name(
                activity_table.c.old_data, old_column_name, new_column_name
            ),
            changed_data=jsonb_change_key_name(
                activity_table.c.changed_data, old_column_name, new_column_name
            ),
        )
        .where(activity_table.c.table_name == table)
    )
    return conn.execute(query)


def add_column(conn, table, column_name, default_value=None, schema=None):
    """
    Adds given column to `activity` table jsonb data columns.

    In the following example we reflect the changes made to our schema to
    activity table.

    ::

        import sqlalchemy as sa
        from alembic import op
        from flask_audit_logger import add_column


        def upgrade():
            op.add_column('article', sa.Column('created_at', sa.DateTime()))
            add_column(op, 'article', 'created_at')


    :param conn:
        An object that is able to execute SQL (either SQLAlchemy Connection,
        Engine or Alembic Operations object)
    :param table:
        The table to remove the column from
    :param column_name:
        Name of the column to add
    :param default_value:
        The default value of the column
    :param schema:
        Optional name of schema to use.
    """
    activity_table = get_activity_table(schema=schema)
    data = {column_name: default_value}
    query = (
        update(activity_table)
        .values(
            old_data=sa.case(
                (
                    sa.cast(activity_table.c.old_data, sa.Text) != "{}",
                    activity_table.c.old_data + data,
                ),
                else_=sa.cast({}, JSONB),
            ),
            changed_data=sa.case(
                (
                    sa.and_(
                        sa.cast(activity_table.c.changed_data, sa.Text) != "{}",
                        activity_table.c.verb != "update",
                    ),
                    activity_table.c.changed_data + data,
                ),
                else_=activity_table.c.changed_data,
            ),
        )
        .where(activity_table.c.table_name == table)
    )
    return conn.execute(query)


def remove_column(conn, table, column_name, schema=None):
    """
    Removes given `activity` jsonb data column key. This function is useful
    when you are doing schema changes that require removing a column.

    Let's say you've been using PostgreSQL-Audit for a while for a table called
    article. Now you want to remove one audited column called 'created_at' from
    this table.

    ::

        from alembic import op
        from flask_audit_logger import remove_column


        def upgrade():
            op.remove_column('article', 'created_at')
            remove_column(op, 'article', 'created_at')


    :param conn:
        An object that is able to execute SQL (either SQLAlchemy Connection,
        Engine or Alembic Operations object)
    :param table:
        The table to remove the column from
    :param column_name:
        Name of the column to remove
    :param schema:
        Optional name of schema to use.
    """
    activity_table = get_activity_table(schema=schema)
    remove = sa.cast(column_name, sa.Text)
    query = (
        update(activity_table)
        .values(
            old_data=activity_table.c.old_data - remove,
            changed_data=activity_table.c.changed_data - remove,
        )
        .where(activity_table.c.table_name == table)
    )
    return conn.execute(query)


def rename_table(conn, old_table_name, new_table_name, schema=None):
    """
    Renames given table in activity table. You should remember to call this
    function whenever you rename a versioned table.

    ::

        from alembic import op
        from flask_audit_logger import rename_table


        def upgrade():
            op.rename_table('article', 'article_v2')
            rename_table(op, 'article', 'article_v2')


    :param conn:
        An object that is able to execute SQL (either SQLAlchemy Connection,
        Engine or Alembic Operations object)
    :param old_table_name:
        The name of table to rename
    :param new_table_name:
        New name of the renamed table
    :param schema:
        Optional name of schema to use.
    """
    activity_table = get_activity_table(schema=schema)
    query = (
        update(activity_table)
        .values(table_name=new_table_name)
        .where(activity_table.c.table_name == old_table_name)
    )
    return conn.execute(query)



from alembic.autogenerate import comparators, renderers
from alembic.operations import MigrateOperation, Operations
from sqlalchemy import text


def setup_schema(audit_logger):
    """These operations ensure the target schema and extensions exist."""

    """ Schema """

    @Operations.register_operation("init_audit_logger_schema")
    class InitAuditLoggerSchemaOp(MigrateOperation):
        @classmethod
        def init_audit_logger_schema(cls, operations):
            op = InitAuditLoggerSchemaOp()
            return operations.invoke(op)

        def reverse(self):
            return RemoveAuditLoggerSchemaOp()

    @Operations.register_operation("remove_audit_logger_schema")
    class RemoveAuditLoggerSchemaOp(MigrateOperation):
        @classmethod
        def remove_audit_logger_schema(cls, operations):
            op = RemoveAuditLoggerSchemaOp()
            return operations.invoke(op)

        def reverse(self):
            return InitAuditLoggerSchemaOp()

    @Operations.implementation_for(InitAuditLoggerSchemaOp)
    def init_audit_logger_schema(operations, operation):
        operations.execute(audit_logger.render_sql_template("create_schema.sql", as_text=False))

    @Operations.implementation_for(RemoveAuditLoggerSchemaOp)
    def remove_audit_logger_schema(operations, operation):
        operations.execute(audit_logger.render_sql_template("drop_schema.sql", as_text=False))

    @renderers.dispatch_for(InitAuditLoggerSchemaOp)
    def render_init_audit_logger_schema(autogen_context, op):
        return "op.init_audit_logger_schema()"

    @renderers.dispatch_for(RemoveAuditLoggerSchemaOp)
    def render_remove_audit_logger_schema(autogen_context, op):
        return "op.remove_audit_logger_schema()"

    @comparators.dispatch_for("schema")
    def check_for_audit_logger_schema(autogen_context, upgrade_ops, schemas):
        schema_exists_sql = text(
            f"SELECT TRUE FROM information_schema.schemata WHERE schema_name = '{audit_logger.schema}'"
        )
        schema_exists = autogen_context.connection.scalar(schema_exists_sql)
        if not schema_exists:
            upgrade_ops.ops.insert(0, InitAuditLoggerSchemaOp())

    """ Extensions """

    @Operations.register_operation("init_audit_logger_extension")
    class InitAuditLoggerExtensionOp(MigrateOperation):
        def __init__(self, extension):
            self.extension = extension

        @classmethod
        def init_audit_logger_extension(cls, operations, extension):
            op = InitAuditLoggerExtensionOp(extension)
            return operations.invoke(op)

        def reverse(self):
            return RemoveAuditLoggerExtensionOp(self.extension)

    @Operations.register_operation("remove_audit_logger_extension")
    class RemoveAuditLoggerExtensionOp(MigrateOperation):
        def __init__(self, extension):
            self.extension = extension

        @classmethod
        def remove_audit_logger_extension(cls, operations, extension):
            op = RemoveAuditLoggerExtensionOp(extension)
            return operations.invoke(op)

        def reverse(self):
            return InitAuditLoggerExtensionOp(self.extension)

    @Operations.implementation_for(InitAuditLoggerExtensionOp)
    def init_audit_logger_extension(operations, operation):
        assert operation.extension == "btree_gist"
        operations.execute(audit_logger.pg_btree_gist_extension.create_sql)

    @Operations.implementation_for(RemoveAuditLoggerExtensionOp)
    def remove_audit_logger_extension(operations, operation):
        assert operation.extension == "btree_gist"
        operations.execute(audit_logger.pg_btree_gist_extension.drop_sql)

    @renderers.dispatch_for(InitAuditLoggerExtensionOp)
    def render_init_audit_logger_extension(autogen_context, op):
        return f"op.init_audit_logger_extension('{op.extension}')"

    @renderers.dispatch_for(RemoveAuditLoggerExtensionOp)
    def render_remove_audit_logger_extension(autogen_context, op):
        return f"op.remove_audit_logger_extension('{op.extension}')"

    @comparators.dispatch_for("schema")
    def check_for_audit_logger_extensions(autogen_context, upgrade_ops, schemas):
        btree_gist = audit_logger.pg_btree_gist_extension.signature
        extension_exists_sql = text(
            f"SELECT TRUE FROM pg_extension WHERE extname = '{btree_gist}'"
        )
        extension_exists = autogen_context.connection.scalar(extension_exists_sql)
        if not extension_exists:
            upgrade_ops.ops.insert(0, InitAuditLoggerExtensionOp(btree_gist))



import json

import pytest
from sqlalchemy import text

from tests.defaults.flask_app import db


@pytest.mark.parametrize(
    ("old", "new", "result"),
    (
        ({}, {}, {}),
        ({"age": 13}, {"age": 14}, {"age": 14}),
        ({"name": "Someone"}, {"name": "Someone"}, {}),
        ({"name": "Someone"}, {"name": "Else"}, {"name": "Else"}),
        ({"ids": [1, 2, 3]}, {"ids": [1, 2]}, {"ids": [1, 2]}),
        ({"ids": [1, 2]}, {"ids": []}, {"ids": []}),
        ({"ids": [1, 2]}, {"ids": [1, 2, 3]}, {"ids": [1, 2, 3]}),
        ({"ids": {}}, {"ids": None}, {"ids": None}),
        ({"ids": None}, {"ids": {}}, {"ids": {}}),
        ({"name": "Someone", "age": 15}, {"name": "Else", "age": 15}, {"name": "Else"}),
    ),
)
def test_jsonb_subtract(test_client, old, new, result):
    assert (
        db.session.scalar(
            text("SELECT jsonb_subtract((:new)::jsonb, (:old)::jsonb)"),
            dict(old=json.dumps(old), new=json.dumps(new)),
        )
        == result
    )



from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

from tests.defaults.flask_app import audit_logger, db

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
# target_metadata = MetaData()
target_metadata = db.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            include_schemas=True,
            process_revision_directives=audit_logger.process_revision_directives,
        )

        with context.begin_transaction():
            context.run_migrations()


run_migrations_online()



from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression


class jsonb_change_key_name(expression.FunctionElement):
    """Provides jsonb_change_key_name as a SQLAlchemy FunctionElement."""

    type = JSONB()
    name = "jsonb_change_key_name"
    inherit_cache = False


@compiles(jsonb_change_key_name)
def compile_jsonb_change_key_name(element, compiler, **kw):
    arg1, arg2, arg3 = list(element.clauses)
    arg1.type = JSONB()
    return "jsonb_change_key_name({0}, {1}, {2})".format(
        compiler.process(arg1), compiler.process(arg2), compiler.process(arg3)
    )



import pytest
from sqlalchemy import select

from tests.custom_actor.flask_app import AuditLogActivity, db


@pytest.mark.usefixtures("test_client")
class TestAuditLoggerCustomActor:
    def test_custom_actor_class(self, superuser, test_client):
        resp = test_client.post("/article")

        activities = db.session.scalars(select(AuditLogActivity)).all()
        activity = activities[0]

        assert resp.status_code == 200
        assert len(activities) == 1
        assert activity.table_name == "article"
        assert activity.verb == "insert"
        assert activity.old_data == {}
        assert activity.data == activity.changed_data == resp.json
        assert activity.transaction.actor_id == superuser.id
        assert activity.transaction.actor == superuser



import os
from datetime import datetime
from typing import Optional

from flask import Flask
from flask_login import LoginManager, UserMixin
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column

from flask_audit_logger import AuditLogger

db_user = os.environ.get("FLASK_AUDIT_LOGGER_TEST_USER", "postgres")
db_password = os.environ.get("FLASK_AUDIT_LOGGER_TEST_PASSWORD", "")
db_name = os.environ.get("FLASK_AUDIT_LOGGER_TEST_DB", "flask_audit_logger_test")
db_conn_str = f"postgresql://{db_user}:{db_password}@localhost/{db_name}"


class Base(MappedAsDataclass, DeclarativeBase):
    pass


db = SQLAlchemy(model_class=Base)
login_manager = LoginManager()


class User(db.Model, UserMixin):
    __tablename__ = "user"
    __table_args__ = ({"info": {"versioned": {"exclude": ["age", "height"]}}},)
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(100))
    age: Mapped[Optional[int]] = mapped_column(default=None)
    height: Mapped[Optional[int]] = mapped_column(default=None)


class Article(db.Model):
    __tablename__ = "article"
    __table_args__ = ({"info": {"versioned": {"exclude": ["created"]}}},)
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    created: Mapped[datetime] = mapped_column(insert_default=datetime.now)


@login_manager.user_loader
def load_user(id):
    return db.session.get(User, id)


app = Flask(__name__)
app.config["SQLALCHEMY_DATABASE_URI"] = db_conn_str
app.secret_key = "secret"
app.debug = True

db.init_app(app)
login_manager.init_app(app)

audit_logger = AuditLogger(db, schema="audit_logs")
AuditLogActivity = audit_logger.activity_cls



from datetime import datetime, timedelta

import pytest
from sqlalchemy import select, text

from tests.separate_schema.flask_app import AuditLogActivity, audit_logger, db


@pytest.mark.usefixtures("test_client")
class TestAuditLoggerSeparateSchema:
    def test_multiple_schemas_exist(self):
        schemas = db.session.scalars(
            text("SELECT schema_name FROM information_schema.schemata")
        ).all()
        assert audit_logger.schema not in [None, "", "public"]
        assert audit_logger.schema in schemas

    def test_audit_log_tables_in_new_schema(self):
        tables = db.session.scalars(
            text(
                f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{audit_logger.schema}'"
            )
        ).all()
        assert sorted(tables) == ["activity", "transaction"]

    def test_one_column_excluded(self, article):
        original_data = {"id": article.id, "name": article.name}
        article.name = "Reba tie a cable to a tree"
        db.session.commit()
        article.created = datetime.now() + timedelta(hours=1)
        db.session.commit()
        activities = db.session.scalars(
            select(AuditLogActivity).order_by(AuditLogActivity.issued_at)
        ).all()
        assert len(activities) == 2

        a0 = activities[0]
        assert a0.verb == "insert"
        assert a0.changed_data == original_data

        a1 = activities[1]
        assert a1.verb == "update"
        assert a1.changed_data == {"name": article.name}

    def test_two_columns_excluded(self, user):
        original_data = {"id": user.id, "name": user.name}

        user.age = 150
        user.height = 999
        db.session.commit()

        activities = db.session.scalars(select(AuditLogActivity)).all()
        assert len(activities) == 1

        a0 = activities[0]
        assert a0.verb == "insert"
        assert a0.changed_data == original_data



import pytest
from sqlalchemy import func, insert, select

from tests.defaults.flask_app import AuditLogActivity, User, db


@pytest.mark.usefixtures("test_client")
class TestAuditLoggerDefaults:
    def test_insert(self, user):
        activity = db.session.scalar(select(AuditLogActivity).limit(1))
        assert activity.old_data == {}
        assert activity.changed_data == {"id": user.id, "name": "Jan", "age": 15}
        assert activity.table_name == "user"
        assert activity.native_transaction_id > 0
        assert activity.verb == "insert"

    def test_activity_after_commit(self):
        user = User(id=1, name="Jack")
        db.session.add(user)
        db.session.commit()
        user = User(id=2, name="Jill")
        db.session.add(user)
        db.session.commit()
        assert db.session.scalar(select(func.count()).select_from(AuditLogActivity)) == 2

    def test_activity_after_rollback(self):
        user = User(id=1, name="Jack")
        db.session.add(user)
        db.session.rollback()
        user = User(id=2, name="Jill")
        db.session.add(user)
        db.session.commit()
        assert db.session.scalar(select(func.count()).select_from(AuditLogActivity)) == 1

    def test_audit_logger_no_actor(self):
        user = User(id=1, name="Jack")
        db.session.add(user)
        db.session.commit()
        activity = db.session.scalar(select(AuditLogActivity).limit(1))
        assert activity.transaction.actor_id is None

    def test_audit_logger_with_actor(self, test_client, logged_in_user):
        resp = test_client.post("/article")
        assert resp.status_code == 200
        activity = db.session.scalar(
            select(AuditLogActivity).where(AuditLogActivity.table_name == "article").limit(1)
        )
        assert activity.transaction.actor_id == logged_in_user.id

    def test_raw_inserts(self):
        db.session.execute(insert(User).values(name="John"))
        db.session.execute(insert(User).values(name="John"))
        activities = db.session.scalars(select(AuditLogActivity)).all()
        assert len(activities) == 2

    def test_activity_repr(self):
        activity = AuditLogActivity()
        activity.id = 3
        activity.table_name = "user"
        assert repr(activity) == "<AuditLogActivity table_name='user' id=3>"

    def test_data_expression_sql(self):
        assert str(AuditLogActivity.data.expression) == (
            "activity.old_data || activity.changed_data"
        )

    def test_data_expression(self, user):
        user.name = "Luke"
        db.session.commit()
        activities = db.session.scalars(
            select(AuditLogActivity).where(
                AuditLogActivity.table_name == "user",
                AuditLogActivity.data["id"].astext == str(user.id),
            )
        ).all()
        assert len(activities) == 2



import os
from dataclasses import asdict
from typing import Optional

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column

from flask_audit_logger import AuditLogger

db_user = os.environ.get("FLASK_AUDIT_LOGGER_TEST_USER", "postgres")
db_password = os.environ.get("FLASK_AUDIT_LOGGER_TEST_PASSWORD", "")
db_name = os.environ.get("FLASK_AUDIT_LOGGER_TEST_DB", "flask_audit_logger_test")
db_conn_str = f"postgresql://{db_user}:{db_password}@localhost/{db_name}"


class Base(MappedAsDataclass, DeclarativeBase):
    pass


db = SQLAlchemy(model_class=Base)


class User(db.Model):
    __tablename__ = "user"
    __table_args__ = ({"info": {"versioned": {}}},)
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(100))
    age: Mapped[Optional[int]] = mapped_column(default=None)


class Article(db.Model):
    __tablename__ = "article"
    __table_args__ = ({"info": {"versioned": {}}},)
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]


class SuperUser(db.Model):
    __tablename__ = "super_user"
    id: Mapped[int] = mapped_column(primary_key=True)


def get_actor_id():
    return 123


app = Flask(__name__)
app.config["SQLALCHEMY_DATABASE_URI"] = db_conn_str
app.secret_key = "secret"
app.debug = True

db.init_app(app)

audit_logger = AuditLogger(db, actor_cls="SuperUser", get_actor_id=get_actor_id)
AuditLogActivity = audit_logger.activity_cls


@app.post("/article")
def create_article():
    article = Article(id=1, name="Down in the west Texas town of El Paso...")
    db.session.add(article)
    db.session.commit()
    return asdict(article)
